/*
 * Copyright (C) 2023  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/libfred/registrable_object/domain/auction/get_auctions.hh"

#include "liblog/liblog.hh"

#include <algorithm>
#include <set>
#include <utility>

namespace LibFred {
namespace Domain {
namespace Auction {

GetAuctions::Result GetAuctions::exec(const OperationContext& ctx, std::initializer_list<Property> filter) const
{
    try
    {
        std::string sql = "SELECT id, "
                                 "COALESCE(external_id, ''), "
                                 "fqdn, "
                                 "external_id IS NULL, "
                                 "next_event_after < NOW() IS NOT TRUE AND winner_id IS NULL AND external_id IS NOT NULL, "
                                 "next_event_after < NOW() AND winner_id IS NULL AND external_id IS NOT NULL, "
                                 "NOW() < win_expires_at AND external_id IS NOT NULL, "
                                 "win_expires_at IS NULL AND winner_id IS NOT NULL AND external_id IS NOT NULL, "
                                 "win_expires_at <= NOW() AND external_id IS NOT NULL "
                            "FROM domain_auction "
                           "WHERE created_at <= NOW() AND "
                                 "finished_at IS NULL";
        if (0 < filter.size())
        {
            const auto properties = std::set<Property>{begin(filter), end(filter)};
            std::string cond;
            std::for_each(begin(properties), end(properties), [&cond](auto&& property)
            {
                const auto extend_or_cond = [&cond](const char* expr)
                {
                    if (!cond.empty())
                    {
                        cond += " OR ";
                    }
                    cond += expr;
                };
                switch (property)
                {
                    case Property::without_external_auction_id:
                        return extend_or_cond("external_id IS NULL");
                    case Property::waiting_for_event:
                        return extend_or_cond("(next_event_after < NOW() IS NOT TRUE AND "
                                               "winner_id IS NULL AND "
                                               "external_id IS NOT NULL)");
                    case Property::event_anticipated:
                        return extend_or_cond("(next_event_after < NOW() AND "
                                               "winner_id IS NULL AND "
                                               "external_id IS NOT NULL)");
                    case Property::waiting_for_registration:
                        return extend_or_cond("(NOW() < win_expires_at AND "
                                               "external_id IS NOT NULL)");
                    case Property::registration_completed:
                        return extend_or_cond("(win_expires_at IS NULL AND "
                                               "winner_id IS NOT NULL AND "
                                               "external_id IS NOT NULL)");
                    case Property::registration_expired:
                        return extend_or_cond("(win_expires_at <= NOW() AND "
                                               "external_id IS NOT NULL)");
                }
                struct UnknownFilter : std::exception
                {
                    const char* what() const noexcept override { return "unknown filter"; }
                };
                throw UnknownFilter{};
            });
            sql += " AND (" + cond + ")";
        }
        sql += " ORDER BY 1";
        const auto dbres = ctx.get_conn().exec(sql);
        Result result;
        result.reserve(dbres.size());
        for (auto idx = 0ull; idx < dbres.size(); ++idx)
        {
            const auto columns = dbres[idx];
            result.emplace_back(
                    AuctionId{static_cast<std::uint64_t>(columns[0])},
                    ExternalAuctionId{static_cast<std::string>(columns[1])},
                    static_cast<std::string>(columns[2]),
                    [&columns]()
                    {
                        if (static_cast<bool>(columns[3]))
                        {
                            return Property::without_external_auction_id;
                        }
                        if (static_cast<bool>(columns[4]))
                        {
                            return Property::waiting_for_event;
                        }
                        if (static_cast<bool>(columns[5]))
                        {
                            return Property::event_anticipated;
                        }
                        if (static_cast<bool>(columns[6]))
                        {
                            return Property::waiting_for_registration;
                        }
                        if (static_cast<bool>(columns[7]))
                        {
                            return Property::registration_completed;
                        }
                        if (static_cast<bool>(columns[8]))
                        {
                            return Property::registration_expired;
                        }
                        struct NoProperty : std::exception
                        {
                            const char* what() const noexcept override { return "no property"; }
                        };
                        throw NoProperty{};
                    }());
        }
        return result;
    }
    catch (const std::exception& e)
    {
        LIBLOG_INFO("std::exception caught: {}", e.what());
        throw;
    }
    catch (...)
    {
        LIBLOG_INFO("unknown exception caught");
        throw;
    }
}

GetAuctions::AuctionInfo::AuctionInfo(
        AuctionId auction_id,
        ExternalAuctionId external_auction_id,
        std::string fqdn,
        Property property)
    : auction_id{auction_id},
      external_auction_id{std::move(external_auction_id)},
      fqdn{std::move(fqdn)},
      property{property}
{ }

}//namespace LibFred::Domain::Auction
}//namespace LibFred::Domain
}//namespace LibFred
